/*****************************************************************************
;
;   (C) Unpublished Work of ADMtek Incorporated.  All Rights Reserved.
;
;       THIS WORK IS AN UNPUBLISHED WORK AND CONTAINS CONFIDENTIAL,
;       PROPRIETARY AND TRADESECRET INFORMATION OF ADMTEK INCORPORATED.
;       ACCESS TO THIS WORK IS RESTRICTED TO (I) ADMTEK EMPLOYEES WHO HAVE A
;       NEED TO KNOW TO PERFORM TASKS WITHIN THE SCOPE OF THEIR ASSIGNMENTS
;       AND (II) ENTITIES OTHER THAN ADMTEK WHO HAVE ENTERED INTO APPROPRIATE
;       LICENSE AGREEMENTS.  NO PART OF THIS WORK MAY BE USED, PRACTICED,
;       PERFORMED, COPIED, DISTRIBUTED, REVISED, MODIFIED, TRANSLATED,
;       ABBRIDGED, CONDENSED, EXPANDED, COLLECTED, COMPILED, LINKED, RECAST,
;       TRANSFORMED OR ADAPTED WITHOUT THE PRIOR WRITTEN CONSENT OF ADMTEK.
;       ANY USE OR EXPLOITATION OF THIS WORK WITHOUT AUTHORIZATION COULD
;       SUBJECT THE PERPERTRATOR TO CRIMINAL AND CIVIL LIABILITY.
;
;------------------------------------------------------------------------------
;
;    Project :  
;    Creator : 
;    File    : ddns_daemon.c
;    Abstract: DDNS configuration parameters.
;
;   Function List:
;
;Modification History:
;       By        0      Date     Ver.   Modification Description
;       --------------- -------- -----  --------------------------------------
;
;*****************************************************************************/

#include <netmall.h>
#include <knl_sup.h>
#include <inet_sup.h>
#include <file_sup.h>
#include <net/if_dl.h>
#include <netinet/in.h>
#include <sys/filio.h>
#include <ddns.h>
#include <ddns_service.h>
#include <ddns_cfg.h>

#define DDNS_NAME_LEN   255

//*************************
// structures
//*************************

typedef struct
{
	int     op;
	UINT32  server;
	char    hostname[DDNS_NAME_LEN+1];	
    DDNS_INFO_T info;
    void    *pUser;
}
DDNS_CMD_T;


#define DDNS_CMD_STOP        1
#define DDNS_CMD_ADDUSER     2
#define DDNS_CMD_DELUSER     3
#define DDNS_CMD_EXPIRE      4
//*************************
// Variables
//*************************
static  KNL_TASK_ID ddnsd_TaskId;
static  KNL_MSGQ_ID ddns_CmdQ = 0;
static  int rtsock;
static  UINT32 uptime;
struct DDNS_USER_S *ddns_user_info = NULL;
int wan_mode = DDNS_CFG_NONE;
int mode_change = FALSE;
UINT32 global_ip;
/*--------------------------------------------------------------
* ROUTINE NAME - ddns_set_mode
*---------------------------------------------------------------
* FUNCTION: 
*
* INPUT:    
* RETURN:
*       
*       
---------------------------------------------------------------*/
UINT32 ddns_set_mode (int mode)
{
	if(wan_mode != mode)
	{
		mode_change = TRUE;
		wan_mode = mode;	
	}
}

/*--------------------------------------------------------------
* ROUTINE NAME - ddns_IpChange
*---------------------------------------------------------------
* FUNCTION: This routine listen to announcements from the kernel
*
* INPUT:    
* RETURN:
*       
*       
---------------------------------------------------------------*/
static int ddns_IpChange()
{
    char msg[512];
    int len;
    struct ifa_msghdr *ifm;
    char *cp; 
    struct sockaddr_dl *dl;
    struct sockaddr_in *addr;
    int  i;
    char ifname[30];
    DDNS_USER_T *pUser;
	
  
    while(1)
    {
    	
    	pUser = ddns_user_info;
    	
        len = read(rtsock, msg, sizeof(msg));
        if (len <= 0) 
        {
            return OK;
        }
        ifm = (struct ifa_msghdr *)msg;
        if (len < ifm->ifam_msglen) 
        {
            printf ("read(PF_ROUTE) short read\n");
            return ERROR;
        }
        
        if (ifm->ifam_version != RTM_VERSION) 
        {
            printf ("routing socket version mismatch\n");
            close(rtsock);
            rtsock = 0;
            return ERROR;
        }
        cp = (char *)(ifm+1);
        
        if (ifm->ifam_type == RTM_NEWADDR &&
            ifm->ifam_addrs & (RTA_IFP|RTA_IFA)) 
        {
            for (i=0; i<RTAX_MAX; i++)
            {
                if (i == RTAX_IFP)
                {
                    dl = (struct sockaddr_dl *)cp;
                    // Print if_name;
                    memcpy(ifname, (char *)dl->sdl_data, dl->sdl_nlen);
                    ifname[dl->sdl_nlen] = 0;
                }
                else if(i == RTAX_IFA)
                {
                    addr = (struct sockaddr_in *)cp;
                }
                
                // Shift cp pointer
                if (ifm->ifam_addrs & (1 << i))
                {
                    cp = (char *)((unsigned long)(cp + (*cp? (*cp+3) : 4)) & ~3);  
                }
            }
        }
        else
            return ERROR;
            
        if(wan_mode == DDNS_CFG_WAN)
        {
	        while (pUser)
	        {	
	            printf("The inteface name is %s", ifname);
	            printf("The puser interface is %s", pUser->info.ifname);
	            if (strcmp(pUser->info.ifname, ifname) == 0)
	            {
	                printf("find the interface\n");
	                if (ifm->ifam_type == RTM_NEWADDR)
	                {
	                    pUser->ipAddr = ntohl(addr->sin_addr.s_addr);
	                    printf("IP is changed");
	                    KNL_SLEEP(300);
	                    ddns_account_update(pUser);
	                }
	    		}
	    		 pUser = pUser->next;  
	        }
	    }        	
    } // end while
    return OK;
}


//##################################################
// EXTERNAL CALLING FUNCTIONS
//##################################################

/*--------------------------------------------------------------
* ROUTINE NAME - ddns_AddUser
*---------------------------------------------------------------
* FUNCTION: 
*
* INPUT:
* RETURN:
*       0   -- no error
---------------------------------------------------------------*/
INT32   ddns_AddUser(DDNS_INFO_T *pInfo)
{
    DDNS_CMD_T cmd;
    
    cmd.op = DDNS_CMD_ADDUSER;
    memcpy(&cmd.info, pInfo, sizeof(DDNS_INFO_T));
    
    return KNL_MSGQ_SEND(ddns_CmdQ, &cmd, sizeof(cmd), KNL_WAIT);
}


/*--------------------------------------------------------------
* ROUTINE NAME - ddns_DelUser
*---------------------------------------------------------------
* FUNCTION: 
*
* INPUT:
* RETURN:
*       0   -- no error
---------------------------------------------------------------*/
INT32   ddns_DelUser(DDNS_INFO_T *pInfo)
{
    DDNS_CMD_T cmd;
    
    cmd.op = DDNS_CMD_DELUSER;
    memcpy(&cmd.info, pInfo, sizeof(DDNS_INFO_T));
    
    return KNL_MSGQ_SEND(ddns_CmdQ, &cmd, sizeof(cmd), KNL_WAIT);
}


/*--------------------------------------------------------------
* ROUTINE NAME - ddns_Stop
*---------------------------------------------------------------
* FUNCTION: 
*
* INPUT:
* RETURN:
*       0   -- no error
---------------------------------------------------------------*/
INT32   ddns_Stop()
{
    DDNS_CMD_T cmd;

    cmd.op = DDNS_CMD_STOP;

    KNL_MSGQ_SEND(ddns_CmdQ, &cmd, sizeof(cmd), KNL_WAIT);

    // Wait until shutdown completed
    while (ddns_CmdQ != 0)
        KNL_SLEEP(1);

    return OK;
}

/*--------------------------------------------------------------*/
/* ROUTINE NAME - ddns_expire    							    */
/*--------------------------------------------------------------*/
/* FUNCTION:													*/
/* INPUT: 														*/
/*--------------------------------------------------------------*/
int ddns_expire(DDNS_USER_T *pUser)
{
    DDNS_CMD_T cmd;

    if(pUser->status == UPDATERES_RETRY)
    {
        cmd.op = DDNS_CMD_EXPIRE;
        cmd.pUser = (void *)pUser;
        KNL_MSGQ_SEND(ddns_CmdQ, &cmd, sizeof(cmd), KNL_NOWAIT);
    }
    return OK;        
}

/*----------------------------------------------------------------------
* ROUTINE NAME - ddns_ProcessModeChange
*-----------------------------------------------------------------------
* DESCRIPTION : 
*
* INPUT 	  : 
* OUPUT 	  : 
* RETURN 	  : 
*----------------------------------------------------------------------*/
static  int ddns_ProcessModeChange()
{
	UINT32 ipAddr=0;	
	DDNS_USER_T *pUser;
	
	KNL_SLEEP(25);
	
	if((KNL_TICKS() - uptime) > 20 && global_ip != 0)
	{
		pUser = ddns_user_info;
		if(wan_mode == DDNS_CFG_PPPOE)
		{
			if (Ifctl_IfIpAddr(&ipAddr, "pppoe0") != ERROR)
			{
				if(global_ip != ipAddr)
				{
					while (pUser)
					{
    					pUser->ipAddr = ipAddr;
    					KNL_SLEEP(300);
    					ddns_account_update(pUser);
				
						pUser = pUser->next;            
					}	
					global_ip = ipAddr;
				}
				
			}
		}	
		else if(wan_mode == DDNS_CFG_PPTP)
		{
			if (Ifctl_IfIpAddr(&ipAddr, "pptp0") != ERROR)
			{
				if(global_ip != ipAddr)
				{
					while (pUser)
					{
    					pUser->ipAddr = ipAddr;
    					KNL_SLEEP(300);
    					ddns_account_update(pUser);
				
						pUser = pUser->next;            
					}	
					global_ip = ipAddr;
				}
				
			}
		}					
		uptime = KNL_TICKS();
	}
	
	
	if (mode_change == TRUE && ddns_user_info != NULL)
	{
		pUser = ddns_user_info;
		
		if(wan_mode == DDNS_CFG_WAN)
		{
			if (Ifctl_IfIpAddr(&ipAddr, "adm1") == ERROR)
			{
				ipAddr = 0;
			}	
		}
		else if(wan_mode == DDNS_CFG_PPPOE)
		{
			if (Ifctl_IfIpAddr(&ipAddr, "pppoe0") == ERROR)
			{
				ipAddr = 0;
			}		
		}
		else if(wan_mode == DDNS_CFG_PPTP)
		{
			if (Ifctl_IfIpAddr(&ipAddr, "pptp0") == ERROR)
			{
				ipAddr = 0;
			}		
		}
		
		if(ipAddr != 0)
		{
			while (pUser)
			{
    			pUser->ipAddr = ipAddr;
    			KNL_SLEEP(300);
    			ddns_account_update(pUser);
				
				pUser = pUser->next;            
			}
			
			mode_change = FALSE;
			global_ip = ipAddr;
			
		}	
	}
		
	return 0;
}
/*----------------------------------------------------------------------
* ROUTINE NAME - ddns_ProcessPkt
*-----------------------------------------------------------------------
* DESCRIPTION : 
*
* INPUT 	  : 
* OUPUT 	  : 
* RETURN 	  : 
*----------------------------------------------------------------------*/
static  int ddns_ProcessPkt()
{   
	int s;
    int n;
    fd_set  fds;
    struct  timeval w;

    //****************
    // Do init
    //****************
    w.tv_sec = 0;
    w.tv_usec = 20000;          /* timed out every 20 ms */

	FD_ZERO(&fds);

    // Set routing socket for IP change
    FD_SET(rtsock, &fds);
	
    n = select(FD_SETSIZE, &fds, (fd_set *)NULL, (fd_set *)NULL, &w);
    if (n > 0)
    {
        // Process IP change
        if (FD_ISSET(rtsock, &fds))
            ddns_IpChange();
    }

    return 0;
}

/*--------------------------------------------------------------
* ROUTINE NAME - ddns_shutdown
*---------------------------------------------------------------
* FUNCTION: This routine delete an interface to the upnpd iflist
*
* INPUT:    interface name
* RETURN:
*       0   -- no error
*       others -- cannot be added
---------------------------------------------------------------*/
static int ddns_shutdown()
{
    DDNS_USER_T *pTemp;
    
     while (ddns_user_info)  
    {
    	pTemp = ddns_user_info;
    	ddns_user_info = ddns_user_info->next;
			
		free(pTemp);
    }         
    
    // Close routing socket
    if (rtsock != -1)
        close(rtsock);
    
    // Delete interface command queue
    KNL_MSGQ_DELETE(ddns_CmdQ);
    ddns_CmdQ = 0;   
    
    //terminate the task         
    KNL_TASK_EXIT();
}         

/*--------------------------------------------------------------
* ROUTINE NAME - ddns_add_client
*---------------------------------------------------------------
* FUNCTION: This routine add the new client information to ddns list
*
* INPUT:    new user information
* RETURN:
*       0   -- no error
*       others -- cannot be added
---------------------------------------------------------------*/
static int ddns_add_client(DDNS_INFO_T *pInfo)
{
    UINT32 ipAddr=0;
    DDNS_USER_T *pUser;

    pUser = ddns_user_info;
    
    // Search for the one already in
    while (pUser)
    {
        if (pUser->info.server == pInfo->server &&
            strcmp(pUser->info.ddns_hostname, pInfo->ddns_hostname) == 0)
            return;

        pUser = pUser->next;            
    }
    
    // Add a new one
    pUser = (DDNS_USER_T *)calloc(1, sizeof(DDNS_USER_T));
    if (pUser == 0)
        return NULL;    
    
    pUser->status = UPDATERES_OK;
    memcpy(&(pUser->info), pInfo, sizeof(DDNS_INFO_T));

    // Do prepend
    pUser->next = ddns_user_info;
    ddns_user_info = pUser;

    mode_change = TRUE;           
    //IpCfg_IpAddr(&ipAddr, pUser->info.ifname);
        
    //if(ipAddr!=0)
    //{
    //    pUser->ipAddr = ipAddr;
    //    ddns_account_update(pUser);
    //}    

    return;    
}         

/*--------------------------------------------------------------
* ROUTINE NAME - ddns_del_client
*---------------------------------------------------------------
* FUNCTION: This routine add the new client information to ddns list
*
* INPUT:    new user information
* RETURN:
*       0   -- no error
*       others -- cannot be added
---------------------------------------------------------------*/
static int ddns_del_client(DDNS_INFO_T *pInfo)
{

    DDNS_USER_T *pUser;
    DDNS_USER_T *pUser_prev;
    
    pUser = ddns_user_info;
    
    
    // Search for the one to delete
    while (pUser)
    {
        if (pUser->info.server == pInfo->server &&
            strcmp(pUser->info.ddns_hostname, pInfo->ddns_hostname) == 0)
            break;
        
        pUser_prev = pUser;
        pUser = pUser->next;            
    }
    
    if (!pUser)
        return -1;
    
    if(pUser->status == UPDATERES_RETRY)
    {
        untimeout((FUNCPTR)ddns_expire, pUser);
    }  
    
    if (pUser_prev == 0)
        ddns_user_info = pUser->next;       
    else
        pUser_prev->next = pUser->next;
    
    free(pUser);     
    
    return 0;
}         


/*----------------------------------------------------------------------
* ROUTINE NAME - ddns_ProcessMsg
*-----------------------------------------------------------------------
* DESCRIPTION : 
*
* INPUT 	  : 
* OUPUT 	  : 
* RETURN 	  : 
*----------------------------------------------------------------------*/
static  int ddns_ProcessMsg()
{
    DDNS_CMD_T cmd;
    
    if (KNL_MSGQ_RECV(ddns_CmdQ, &cmd, sizeof(cmd), KNL_NOWAIT) != ERROR)
    {
        switch (cmd.op)
        { 
        case DDNS_CMD_ADDUSER:
            ddns_add_client(&cmd.info);
            break;
            
        case DDNS_CMD_DELUSER:
            ddns_del_client(&cmd.info);
            break;
            
        case DDNS_CMD_STOP:
            ddns_shutdown();
            break;
            
        case DDNS_CMD_EXPIRE:
            ddns_account_update((DDNS_USER_T *)cmd.pUser);
            break;
            
        default:
            break;
        }
    }
    
    return 0;
}



/*--------------------------------------------------------------
* ROUTINE NAME - ddnsd
*---------------------------------------------------------------
* FUNCTION: This is the deamon routine of DDNS.
*           We do some initialization and then enter the
*           main loop to process the following things
*
*           1. Management message
*           2. DDNS packets and interface change message (rtsock)
*         
* INPUT:
* OUTPUT:
* RETURN:
*---------------------------------------------------------------*/
void    ddnsd()
{
	int s;
    fd_set  fds;

    struct  timeval w;
    int n;
    int on;
    DDNS_CMD_T cmd;
    
	
    rtsock = socket(AF_ROUTE, SOCK_RAW, 0);
	
	if (rtsock < 0)
    {
	    printf("\nddnsd_Daemon:: rt_socket create failed");
	    ddns_shutdown();
        return;
    }
    
    on = 1;
    
    if (ioctl(rtsock, FIONBIO, (int) &on) == -1)
    {
	    printf("\nddnsd_Daemon:: ioctl failed");
	    ddns_shutdown();
        return;
    }

	on = 0;
	if (setsockopt(rtsock, SOL_SOCKET,SO_USELOOPBACK, &on,sizeof(on)) < 0)
    {
  	    printf("\nddnsd_Daemon:: setsockopt failed");
        return;
    }


    w.tv_sec = 0;
    w.tv_usec = 50000;  /* timed out every 50 ms */
       
 
    uptime = KNL_TICKS();
    global_ip = 0;
    /*************************************
    loop forever waiting for requests
    **************************************/
    while(1) 
    {
        // Process commands
        ddns_ProcessMsg();

        //Process update for mode change
        ddns_ProcessModeChange ();
        
        // Process Packet
        ddns_ProcessPkt();
    }
}


/*--------------------------------------------------------------
* ROUTINE NAME - ddns_Start
*---------------------------------------------------------------
* FUNCTION: This routine creates the ddns deamon
*
* INPUT:
* RETURN:
*       0   -- no error
*       others -- cannot start
---------------------------------------------------------------*/
INT32   ddns_Start()
{
	int i;

    if (ddns_CmdQ != 0)
        return OK;

    //
    // Create message queue
    //
    if (KNL_MSGQ_CREATE("ddnsq", &ddns_CmdQ, 1, sizeof(DDNS_CMD_T)) != OK)
        return ERROR;

    //
    // Spawn Ddns daemon task
    //
    KNL_TASK_CREATE("tDdnsD", &ddnsd_TaskId, 0, (FUNCPTR)ddnsd, 0, 0); 
    
	return OK;
}

